/*
 * Copyright (C) 2021, Soren Friis
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Open Drone ID Linux transmitter example.
 *
 * Maintainer: Soren Friis
 * friissoren2@gmail.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>
#include <semaphore.h>
#include <signal.h>
#include <sys/resource.h>
#include "ap_interface.h"
#include "bluetooth.h"
#include "wifi_beacon.h"
#include "gpsmod.h"

#include <assert.h>
#include <drip.h>
#include <errno.h>
#include <utils.h>

sem_t semaphore;
pthread_t id, gps_thread;

#define MINIMUM(a,b) (((a)<(b))?(a):(b))

#define BASIC_ID_POS_ZERO 0
#define BASIC_ID_POS_ONE 1

static struct config_data config = { 0 };
static bool kill_program = false;

static struct fixsource_t source;
static struct gps_data_t gpsdata;

struct gps_loop_args {
    struct gps_data_t *gpsdata;
    struct ODID_UAS_Data *uasData;
    int exit_status;
};

static char identities_path[512];

static void fill_example_data(struct ODID_UAS_Data *uasData) {
    uasData->System.OperatorLocationType = ODID_OPERATOR_LOCATION_TYPE_TAKEOFF;
    uasData->System.ClassificationType = ODID_CLASSIFICATION_TYPE_EU;
    uasData->System.OperatorLatitude = uasData->Location.Latitude + 0.001;
    uasData->System.OperatorLongitude = uasData->Location.Longitude - 0.001;
    uasData->System.AreaCount = 1;
    uasData->System.AreaRadius = 0;
    uasData->System.AreaCeiling = 0;
    uasData->System.AreaFloor = 0;
    uasData->System.CategoryEU = ODID_CATEGORY_EU_OPEN;
    uasData->System.ClassEU = ODID_CLASS_EU_CLASS_1;
    uasData->System.OperatorAltitudeGeo = 20.5f;
    uasData->System.Timestamp = 28056789;

    uasData->OperatorID.OperatorIdType = ODID_OPERATOR_ID;
    char operatorId[] = "FIN87astrdge12k8";
    strncpy(uasData->OperatorID.OperatorId, operatorId,
            MINIMUM(sizeof(operatorId), sizeof(uasData->OperatorID.OperatorId)));
}

static void fill_example_gps_data(struct ODID_UAS_Data *uasData) {
    uasData->Location.Status = ODID_STATUS_AIRBORNE;
    uasData->Location.Direction = 361.f;
    uasData->Location.SpeedHorizontal = 0.0f;
    uasData->Location.SpeedVertical = 0.35f;
    uasData->Location.Latitude = 51.4791;
    uasData->Location.Longitude = -0.0013;
    uasData->Location.AltitudeBaro = 100;
    uasData->Location.AltitudeGeo = 110;
    uasData->Location.HeightType = ODID_HEIGHT_REF_OVER_GROUND;
    uasData->Location.Height = 80;
    uasData->Location.HorizAccuracy = createEnumHorizontalAccuracy(5.5f);
    uasData->Location.VertAccuracy = createEnumVerticalAccuracy(9.5f);
    uasData->Location.BaroAccuracy = createEnumVerticalAccuracy(0.5f);
    uasData->Location.SpeedAccuracy = createEnumSpeedAccuracy(0.5f);
    uasData->Location.TSAccuracy = createEnumTimestampAccuracy(0.1f);
    uasData->Location.TimeStamp = 360.52f;
}

static void cleanup(int exit_code) {
    if (config.use_btl || config.use_bt4 || config.use_bt5)
        close_bluetooth(&config);

    if (config.use_beacon) {
        send_quit();

        int *ptr;
        pthread_join(id, (void **) &ptr);
        printf("Return value from ap_interface_init: %i\n", *ptr);

        sem_destroy(&semaphore);
    }

    if(config.use_gps) {
        int *ptr;
        pthread_join(gps_thread, (void **) &ptr);
        printf("Return value from gps_loop: %d\n", *ptr);

        gps_close(&gpsdata);
    }

    exit(exit_code);
}

static void sig_handler(int signo) {
    if (signo == SIGINT || signo == SIGSTOP || signo == SIGKILL || signo == SIGTERM) {
        kill_program = true;
    }
}

static void send_message(union ODID_Message_encoded *encoded, struct config_data *config, uint8_t msg_counter) {
    if (config->use_btl)
        send_bluetooth_message(encoded, msg_counter, config);
    if (config->use_bt4 || config->use_bt5)
        send_bluetooth_message_extended_api(encoded, msg_counter, config);
    if (config->use_beacon)
        send_beacon_message(encoded, msg_counter);
}

static void send_packs(ODID_Message_encoded *encoded_msgs,uint8_t count,struct config_data *config){
    struct ODID_MessagePack_encoded pack_enc = { 0 };
	struct ODID_MessagePack_data pack_data = { 0 };
    pack_data.MsgPackSize = count;
	pack_data.SingleMessageSize = ODID_MESSAGE_SIZE;
    for(int i = 0; i<count; i++)
		memcpy(&pack_data.Messages[i], &encoded_msgs[i], ODID_MESSAGE_SIZE);

	if(encodeMessagePack(&pack_enc, &pack_data) != ODID_SUCCESS)
		printf("Failed to encode message pack\n");
	else
	{
		send_beacon_message_pack(&pack_enc, config->msg_counters[ODID_MSG_COUNTER_PACKED]++);
	}
    sleep(4);
}

void print_help() {
    printf("Program for transmitting static drone ID data on Wi-Fi Beacon or Bluetooth.\n");
    printf("Must be run with sudo rights in order to work.\n");
    printf("Options: b Enable Wi-Fi Beacon transmission\n");
    printf("         l Enable Bluetooth 4 Legacy Advertising transmission\n");
    printf("           using the non-Extended Advertising HCI API commands\n");
    printf("         4 Enable Bluetooth 4 Legacy Advertising transmission\n");
    printf("           using the Extended Advertising HCI API commands\n");
    printf("         5 Enable Bluetooth 5 Long Range + Extended Advertising transmission\n");
    printf("         p Use message packs instead of single messages\n");
    printf("         g Use gpsd to dynamically update location messages after each loop of messages\n");
    printf("E.g. sudo ./transmit b p\n\n");
    printf("Wi-Fi Beacon transmit only works when running\n");
    printf("\"sudo hostapd/hostapd/hostapd beacon.conf\" in a separate shell.\n");
    printf("Disconnect from all Wi-Fi networks before starting Wi-Fi Beacon transmission.\n\n");
    printf("If terminated abnormally, Beacon and Bluetooth broadcasts can remain on.\n");
    printf(" - To stop Beacon broadcast, stop the hostapd instance.\n");
    printf("   It can be difficult to stop the transmit instance.\n");
    printf("   After stopping hostapd, use \"sudo pkill transmit\".\n");
    printf(" - To stop Bluetooth, use \"sudo btmgmt power off\" and then\n");
    printf("   \"sudo btmgmt power on\".\n");
}

static void parse_command_line(int argc, char *argv[], struct config_data *config) {
    if (argc == 1) {
        print_help();
        exit(EXIT_SUCCESS);
    }

    if(!strcmp(argv[1], "--identities-path")) {
        int p_len = strlen(argv[2]);
        if(p_len<512)
            strncpy(identities_path, argv[2], p_len);
        else {
            printf("--identities-path arg is too long\n");
            exit(EXIT_FAILURE);
        }
    }
    else {
        printf("--identities-path arg not specified\n");
        exit(EXIT_FAILURE);
    }

    for (int i = 3; i < argc; i++) {
        switch (*argv[i]) {
            case 'b':
                config->use_beacon = true;
                break;
            case 'l':
                config->use_btl = true;
                break;
            case '4':
                config->use_bt4 = true;
                break;
            case '5':
                config->use_bt5 = true;
                break;
            case 'p':
                config->use_packs = true;
                break;
            case 'g':
                config->use_gps = true;
                break;
            default:
                break;
        }
    }
    if (config->use_beacon)
        printf("\nReminder: Wi-Fi Beacon only works when running\n\"sudo hostapd/hostapd/hostapd beacon.conf\" in a separate shell.\n\n");
    if (config->use_beacon && !config->use_packs)
        printf("\nWarning: Transmitting single messages on Wi-Fi beacon is violating\nthe standards. Enable message packs.\n\n");

    if (config->use_btl && (config->use_bt4 || config->use_bt5)) {
        printf("\nError: Cannot use both old API and Extended Advertising API at the same time.\n\n");
        exit(EXIT_FAILURE);
    }
    if ((config->use_btl || config->use_bt4) && config->use_packs) {
        printf("\nError: BT4 cannot use message packs.\n\n");
        exit(EXIT_FAILURE);
    }
    if (config->use_bt4 && config->use_bt5)
        printf("\nWarning: Doing simultaneous BT4 and BT5 will not necessarily work.\n\n");
    if (config->use_bt5 && !config->use_packs)
        printf("\nWarning: Transmitting single messages on Bluetooth 5 Long Range is violating\nthe standards. Enable message packs.\n\n");

    if (!config->use_beacon && !config->use_btl && !config->use_bt4 && !config->use_bt5) {
        print_help();
        exit(EXIT_SUCCESS);
    }

    if (config->use_gps)
        printf("\nWarning: Fetching GPS data requires a configured GPS sensor.\n\n");
}

void gps_loop(struct gps_loop_args *args) {
    struct gps_data_t *gpsdata = args->gpsdata;
    struct ODID_UAS_Data *uasData = args->uasData;

    char gpsd_message[GPS_JSON_RESPONSE_MAX];
    int retries = 0;      // cycles to wait before gpsd timeout
    int read_retries = 0;
    while(true) {
        if(kill_program)
            break;

        int ret;
        ret = gps_waiting(gpsdata, GPS_WAIT_TIME_MICROSECS);
        if (!ret) {
            printf("Socket not ready, retrying...\n");
            if (retries++ > MAX_GPS_WAIT_RETRIES) {
                    fprintf(stderr, "Max socket wait retries reached, exiting...");
                kill_program = true;
                args->exit_status = 1;
                pthread_exit((void*) &args->exit_status);
            }
            continue;
        } else {
            retries = 0;
            gpsd_message[0] = '\0';

            if (gps_read(gpsdata, gpsd_message, sizeof(gpsd_message)) == -1) {
                printf("Failed to read from socket, retrying...\n");
                if(read_retries++ > MAX_GPS_READ_RETRIES) {
                    fprintf(stderr, "Max socket read retries reached, exiting...");
                    kill_program = true;
                    args->exit_status = 1;
                    pthread_exit((void*) &args->exit_status);
                }
                continue;
            }
            read_retries = 0;

            process_gps_data(gpsdata, uasData);
        }
    }

    args->exit_status = 0;
    pthread_exit(&args->exit_status);
}

static void send_encoded_messages(ODID_Message_encoded* encoded, struct config_data* config, short pages, size_t *counter)
{
    size_t mock = 0;
    if(counter == NULL)
        counter = &mock;

    if(!config->use_packs) {
        for(short i = 0; i < pages; i++)
        {
            send_message(&encoded[i], config, (*counter)++);
            sleep(1);
        }
    }
}

static time_t astm_time() {
    return time(NULL);
}


int main(int argc, char *argv[])
{
    parse_command_line(argc, argv, &config);

    DRIP_Core_Config drip_core_config;
    drip_core_config.legacy_transport = config.use_btl;

    DRIP_SetConfig(&drip_core_config);
    DRIP_SetSignDataFun(sign_data);
    DRIP_SetASTMDateTimeFun(astm_time);

    DRIP_HostIdentity *drone = DRIP_HostIdentityFromFile(identities_path,"drone");
    //Two endorsements since we don't have apex level so far 
    //And we don't know the pbkey of RAA so there is no way to verify the endorsement signature in receiver side.
    // If all above requirements meet, it should be 4 endorsements in total(drip-dki-02 section 1.)
    const char endorsement_ua_ihda[DRIP_ENDORSEMENT_LEN*2+1];
    const char endorsement_ihda_ahda[DRIP_ENDORSEMENT_LEN*2+1];


    DRIP_EndorsementFromFile(endorsement_ua_ihda,identities_path,"drone");
    DRIP_EndorsementFromFile(endorsement_ihda_ahda,identities_path,"ihda");

    assert(drone->is_hhit_verified);
    printf("Identity drone read: %p \n", drone->serial_number);

    config.handle_bt4 = 0; // The Extended Advertising set number used for BT4
    config.handle_bt5 = 1; // The Extended Advertising set number used for BT5

    if (config.use_beacon) {
        sem_init(&semaphore,0,0);
        pthread_create(&id, NULL, ap_interface_init, NULL);
        sem_wait(&semaphore);
    }

    struct ODID_UAS_Data uasData;
    odid_initUasData(&uasData);
    fill_example_data(&uasData);
    if(!config.use_gps)
        fill_example_gps_data(&uasData);

    if (config.use_btl || config.use_bt4 || config.use_bt5)
        init_bluetooth(&config);

    size_t location_counter = 0;
    size_t drip_link_1_counter = 0;
    size_t drip_link_2_counter = 0;
    size_t drip_manifest_counter = 0;
    size_t drip_wrapper_counter = 0;

    unsigned char manifest_hash[DRIP_HASH_LEN] = {0};
    const char *initial_manifest = "nonce";
    //ignore \n
    cshake128_wrapper(manifest_hash, (uint8_t *)initial_manifest, strlen(initial_manifest) - 1);

    if(config.use_gps) {
        signal(SIGINT,  sig_handler);
        signal(SIGKILL, sig_handler);
        signal(SIGSTOP, sig_handler);
        signal(SIGTERM, sig_handler);

        if(init_gps(&source, &gpsdata) != 0) {
            fprintf(stderr,
                    "No gpsd running or network error: %d, %s\n",
                    errno, gps_errstr(errno));
            cleanup(EXIT_FAILURE);
        }

        struct gps_loop_args args;
        args.gpsdata = &gpsdata;
        args.uasData = &uasData;
        pthread_create(&gps_thread, NULL, (void*) &gps_loop, &args);

    }

    ODID_Message_encoded encoded[6];

    uasData.SelfID.DescType = ODID_DESC_TYPE_TEXT;
    strcpy(uasData.SelfID.Desc,  "DRIP_SELFID_TEST");

    ODID_BasicID_data basic_id_1 = {0};
    basic_id_1.IDType = ODID_IDTYPE_SPECIFIC_SESSION_ID;
    basic_id_1.UAType = ODID_UATYPE_HELICOPTER_OR_MULTIROTOR;
    memcpy(basic_id_1.UASID, drone->hhit, DRIP_HHIT_LEN);

    ODID_BasicID_data basic_id_2 = {0};
    basic_id_2.IDType = ODID_IDTYPE_SERIAL_NUMBER;
    basic_id_2.UAType = ODID_UATYPE_HELICOPTER_OR_MULTIROTOR;
    memcpy(basic_id_2.UASID, drone->serial_number, DRIP_SERIAL_NUMBER_LEN);

    //DRIP Link
    ODID_Message_encoded encoded_link_ua_ihda[16];
    ODID_Message_encoded encoded_link_ihda_ahda[16];
    short link_pages;

    if(encodeDripLink((ODID_Auth_encoded*)encoded_link_ua_ihda, endorsement_ua_ihda, &link_pages) != ODID_SUCCESS)
        perror("Failed to encode DripLink (BE: issue_HDA, UA)");
    if(encodeDripLink((ODID_Auth_encoded*)encoded_link_ihda_ahda, endorsement_ihda_ahda, &link_pages) != ODID_SUCCESS)
        perror("Failed to encode DripLink (BE: auth_HDA, issue_HDA)");

    uint8_t drip_link_be_hda_ua_hash[DRIP_HASH_LEN] = {0};
    //currently it's just easier to build the DRIP SAM packet from scratch and hash it rather than extract it from ASTM pages
    DRIP_SAM_Data_Format sam_data = {0};
    sam_data.sam_type = DRIP_LINK;
    size_t data_len;
    //RFC 9575 4.4.2 only _DRIP Link (BE: HDA, UA) Hash_ should be sent
    hex_string_to_bytes(endorsement_ua_ihda, sam_data.sam_auth_data, &data_len);
    //+ 1 to account for the SAM type
    cshake128_wrapper(drip_link_be_hda_ua_hash, (uint8_t *)&sam_data, data_len + 1);

    //static ASTM messages
    if(encodeBasicIDMessage((ODID_BasicID_encoded *)&encoded[0], &basic_id_1) != ODID_SUCCESS)
        perror("Failed to encode BasicID 1");

    if(encodeBasicIDMessage((ODID_BasicID_encoded *)&encoded[1], &basic_id_2) != ODID_SUCCESS)
        perror("Failed to encode BasicID 2");

    if(encodeOperatorIDMessage((ODID_OperatorID_encoded *)&encoded[2], &uasData.OperatorID) != ODID_SUCCESS)
        perror("Failed to encode OperatorID");

    if(encodeSelfIDMessage((ODID_SelfID_encoded *)&encoded[3], &uasData.SelfID) != ODID_SUCCESS)
        perror("Failed to encode SelfID");


    while(true)
    {
        if(kill_program)
            break;

        printf("Sending drip link...\n");
        //TODO reduce the rate for sending ihda_ahda and ahda_raa
        send_encoded_messages(encoded_link_ua_ihda, &config, link_pages, &drip_link_1_counter);
        send_encoded_messages(encoded_link_ihda_ahda, &config, link_pages, &drip_link_2_counter);
       

        printf("Sending basic ID 1\n");
        send_encoded_messages(&encoded[0], &config, 1, NULL);

        printf("Sending basic ID 2\n");
        send_encoded_messages(&encoded[1], &config, 1, NULL);

        printf("Sending OperatorID\n");
        send_encoded_messages(&encoded[2], &config, 1, NULL);

        printf("Sending SelfID\n");
        send_encoded_messages(&encoded[3], &config, 1, NULL);

        if(encodeSystemMessage((ODID_System_encoded *)&encoded[4], &uasData.System) != ODID_SUCCESS)
            perror("Failed to encode System");
        else {
            printf("Sending System\n");
            send_encoded_messages(&encoded[4], &config, 1, NULL);
        }

        //Location message (updated by other thread is gpsd is enabled)
        if(encodeLocationMessage((ODID_Location_encoded *)&encoded[5], &uasData.Location) != ODID_SUCCESS)
            perror("Failed to encode Location");
        else {
            printf("Sending Location\n");
            send_encoded_messages(&encoded[5], &config, 1, &location_counter);
        }

        //DRIP Manifest
        ODID_Message_encoded encoded_manifest[16];
        short manifest_pages;

        //first three messages are "manifested"
        if(encodeDripManifest((ODID_Auth_encoded*)encoded_manifest, drone, encoded, 3, manifest_hash, drip_link_be_hda_ua_hash, manifest_hash, &manifest_pages) != ODID_SUCCESS)
            perror("Failed to encode drip manifest");
        else {
            printf("Sending drip manifest\n");
            send_encoded_messages(encoded_manifest, &config, manifest_pages, &drip_manifest_counter);
        }


        ODID_Message_encoded encoded_wrapper[16];
        short wrapper_pages;
        //remaining messages are "wrapped"
        if(encodeDripWrapper((ODID_Auth_encoded*)encoded_wrapper, drone, &encoded[3], 3, &wrapper_pages) != ODID_SUCCESS)
            perror("Failed to encode Drip Wrapper");
        else {
            printf("Sending Drip Wrapper\n");
            send_encoded_messages(encoded_wrapper, &config, wrapper_pages, &drip_wrapper_counter);
        }

        if(config.use_packs){
            send_packs(encoded_link_ua_ihda,link_pages,&config);
            send_packs(encoded_link_ihda_ahda,link_pages,&config);
            send_packs(encoded,6,&config);
            send_packs(encoded_wrapper,wrapper_pages,&config);
            send_packs(encoded_manifest,manifest_pages,&config);
        }

        //TODO reset other counters as well
        drip_link_1_counter = 0;
        drip_link_2_counter = 0;

        sleep(3);
    }

    cleanup(EXIT_SUCCESS);
    return EXIT_SUCCESS;
}